
// Preteceni:
// a + b > INT_MAX
// b > 0: a > INT_MAX - b
// b < 0: b > INT_MAX - a
// Podteceni:
// a + b < INT_MIN
// b < 0: a < INT_MIN - b
// b > 0: b < INT_MIN - a

// |a|*|b| > INT_MAX
// |a| > INT_MAX / |b|

// b == 0

#include "stdafx.h"
#include "Math.h"
#include <limits.h>

int Math::Secti(int a, int b)
{
	// chyba muze nastat jen pokud jsou znaminka stejna
	// kdyz jsou kladna, tak preteceni,  kdyz zaporna, tak podteceni
	if (a > 0 && b > 0)
	{
		if (a > INT_MAX - b)
			throw OverflowException();
	}
	else if (a < 0 && b < 0)
	{
		if (b < INT_MIN - a)
			throw UnderflowException();
	}

	// ted uz to bude v poradku
	return a + b;
}

int Math::Odecti(int a, int b)
{
	// zaporna cisla maji rozsah o jedna vetsi nez kladna, musime
	// otestovat, jestli b neni to nejmensi zaporne cislo -> pak
	// by neslo zmenit znamenko (preteklo by v kladnych)
	if (b == INT_MIN)
	{
		// kdyz k obema pricteme jednicku, nic se nezmeni
		if (a == INT_MAX)
			throw OverflowException(); // nejde pricist jednicka a stejne by preteklo

		++b;
		++a;
	}

	return Secti(a, -b);
}

int Math::Nasob(int a, int b)
{
	// nuly vyresime zvlast, netreba se s nimi pak zabyvat
	if (b == 0 || a == 0)
		return 0;

	// Jestli pretece nebo podtece zavisi jen na velikosti, vezmeme si absolutni
	// hodnotu - to nemuzeme, pokud je neco INT_MIN
	if (a == INT_MIN)
	{
		if (b > 1)
			throw UnderflowException();
		if (b < 0)
			throw OverflowException();
		return a*b;
	}
	if (b == INT_MIN)
	{
		if (a > 1)
			throw UnderflowException();
		if (a < 0)
			throw OverflowException();
		return a*b;
	}
	
	// absolutni hodnoty
	int abs_a = a < 0 ? -a : a;
	int abs_b = b < 0 ? -b : b;

	if (abs_a > INT_MAX / abs_b) // abs_b nebude nula
	{
		if ((a > 0 && b > 0) || (a < 0 && b < 0))
			throw OverflowException();
		else
			throw UnderflowException();
	}

	// uz vse v poradku
	return a * b;
}

int Math::Vydel(int a, int b)
{
	if (b == 0)
		throw ZeroDivisionException();
	
	return a / b;
}
